"""Defines a SARIF output handler for STACS.

SPDX-License-Identifier: BSD-3-Clause
"""

import base64
import json
import re
from typing import Any, Dict, List, Optional, Tuple

from stacs import __about__, model
from stacs.constants import ARCHIVE_FILE_SEPARATOR

# Only one SARIF version will be supported at a time.
SARIF_VERSION = "2.1.0"
SARIF_SCHEMA_URI = "https://schemastore.azurewebsites.net/schemas/json/sarif-2.1.0.json"
SARIF_URI_BASE_ID = "STACSROOT"


def confidence_to_level(confidence: int) -> str:
    """Maps the confidence of a finding to a SARIF level."""
    if confidence < 70:
        return "warning"
    else:
        return "error"


def render_artifact(path: str, parent: Optional[int] = None) -> Dict[str, Any]:
    """Create a new artifact entry."""
    artifact = {
        "location": {
            "uri": path,
            "uriBaseId": SARIF_URI_BASE_ID,
        },
    }

    if parent is not None:
        artifact["parentIndex"] = parent

    return artifact


def path_in_artifacts(path: str, artifacts: List[Dict[str, Any]], parent) -> int:
    """Checks if a path exists in the artifacts list."""
    for index, artifact in enumerate(artifacts):
        if path == artifact["location"]["uri"]:
            # Short circuit if we both don't have a parent.
            if artifact.get("parentIndex", None) is None and parent is None:
                return index

            # Check common ancestry.
            try:
                their_parent = artifact.get("parentIndex", None)
                our_parent = parent

                while True:
                    if our_parent == their_parent:
                        their_parent = artifacts[their_parent]["parentIndex"]
                        our_parent = artifacts[our_parent]["parentIndex"]
                    else:
                        break
            except KeyError:
                # We're good all the way back to the root.
                return index

    return None


def add_artifact(
    root: str,
    finding: model.finding.Entry,
    artifacts: List[Dict[str, Any]],
) -> Tuple[int, List[Dict[str, Any]]]:
    """Generates SARIF artifact entires for findings (SARIF v2.1.0 Section 3.24)."""
    parent = None

    for real_path in finding.path.split(ARCHIVE_FILE_SEPARATOR):
        # Strip the scan directory root from the path for Base URIs to work properly.
        path = re.sub(rf"^{root}", "", real_path).lstrip("/")

        # Check if the path already exists.
        new_parent = path_in_artifacts(path, artifacts, parent)
        if new_parent is not None:
            parent = new_parent
            continue

        artifacts.append(render_artifact(path, parent))
        parent = len(artifacts) - 1

    # Add metadata to this entry, if missing.
    artifacts[parent]["hashes"] = {
        "md5": finding.md5,
    }
    return (parent, artifacts)


def render(
    root: str, findings: List[model.finding.Entry], pack: model.pack.Format
) -> str:
    """Renders down a SARIF document for STACS findings."""
    rules = []
    results = []
    artifacts = []

    # Generate a result (SARIF v2.1.0 Section 3.27) for each finding.
    for finding in findings:
        # Suppressions (SARIF v2.1.0 Section 3.27.23) are used to track findings where
        # there is an "ignore" set - via ignore list.
        suppressions = []

        # Create an artifactContent (SARIF v2.1.0 Section 3.3) entry to track the sample
        # of the finding.
        context_content = {}
        artifact_content = {}

        if finding.sample.binary:
            artifact_content["binary"] = finding.sample.finding
            # Unencode and then re-encode the sample into a single B64 string to provide
            # context.
            context_content["binary"] = base64.b64encode(
                base64.b64decode(finding.sample.before)
                + base64.b64decode(finding.sample.finding)
                + base64.b64decode(finding.sample.after)
            )
        else:
            artifact_content["text"] = finding.sample.finding
            context_content["text"] = (
                finding.sample.before + finding.sample.finding + finding.sample.after
            )

        # Create a new contextRegion (SARIF v2.1.0 Section 3.29.5) to provide contextual
        # information about the finding, but do not include the byte or line number
        # offset.
        context = {"snippet": context_content}

        # Create a new region (SARIF v2.1.0 Section 3.30) to track the location of the
        # finding and the sample.
        region = {
            "byteOffset": finding.location.offset,
            "snippet": artifact_content,
        }

        # Line numbers are optional, as the input file may be binary.
        if finding.location.line:
            region["startLine"] = finding.location.line

        # Add a new artifact for this finding, or retrieve the location of the existing.
        index, artifacts = add_artifact(root, finding, artifacts)

        # Strip the scan directory root from the path, as the we're using the reference
        # from originalUriBaseIds (SARIF v2.1.0 Section 3.14.14) to allow "portability".
        path = finding.path.split(ARCHIVE_FILE_SEPARATOR)[-1]
        relative_path = re.sub(rf"^{root}", "", path).lstrip("/")

        # Pin the artifact location back to a physical location (SARIF v2.1.0 Section
        # 3.28.3).
        physical_location = {
            "physicalLocation": {
                "region": region,
                "contextRegion": context,
                "artifactLocation": {
                    "uri": relative_path,
                    "index": index,
                    "uriBaseId": SARIF_URI_BASE_ID,
                },
            },
        }

        # Generate a new Rule entry, if required (SARIF v2.1.0 Section 3.49).
        rule = None

        for candidate in rules:
            if finding.source.reference == candidate.get("id"):
                rule = candidate
                break

        if not rule:
            # Add the description from the original rule pack entry into the Rule for
            # easy tracking.
            rule = {
                "id": finding.source.reference,
                "shortDescription": {
                    "text": finding.source.description,
                },
            }
            rules.append(rule)

        # Add a Suppression entry if this finding was marked as "Ignored", along with
        # the reason (justification) from the original ignore list.
        if finding.ignore is not None and finding.ignore.ignored:
            suppressions.append(
                {
                    "kind": "external",
                    "status": "accepted",
                    "justification": finding.ignore.reason,
                }
            )

        # Track the finding (Result).
        results.append(
            {
                "message": rule.get("shortDescription"),
                "level": confidence_to_level(finding.confidence),
                "ruleId": finding.source.reference,
                "locations": [
                    physical_location,
                ],
                "suppressions": suppressions,
            }
        )

    # Add a toolComponent (SARIF v2.1.0 Section 3.19), and bolt it all together.
    tool = {
        "driver": {
            "name": __about__.__title__.upper(),
            "rules": rules,
            "version": __about__.__version__,
            "downloadUri": __about__.__uri__,
            "informationUri": __about__.__uri__,
        },
    }
    run = {
        "tool": tool,
        "results": results,
        "artifacts": artifacts,
        "originalUriBaseIds": {
            SARIF_URI_BASE_ID: {
                "uri": f"file://{root.rstrip('/')}/",
            },
        },
    }
    sarif = {
        "version": SARIF_VERSION,
        "$schema": SARIF_SCHEMA_URI,
        "runs": [
            run,
        ],
    }

    # Return a stringified JSON representation of the SARIF document.
    return json.dumps(sarif)



"""Defines a manifest loader for STACS.

SPDX-License-Identifier: BSD-3-Clause
"""

#
# TODO: Implement the manifest loader. This should take the contents of a manifest
#       which matches the stacs.model.manifest.Format schema. It should also check
#       whether all requested files exist, and generate MD5 sums for them if not
#       specified in the manifest 'Entry' (stacs.model.manifest.Entry).
#



"""Implements a rules based scanner for STACS.

SPDX-License-Identifier: BSD-3-Clause
"""

import base64
import hashlib
import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List

import yara
from stacs.constants import CHUNK_SIZE, WINDOW_SIZE
from stacs.exceptions import FileAccessException, InvalidFormatException
from stacs.loader import archive
from stacs.model import finding, manifest, pack


def is_binary(target: manifest.Entry) -> bool:
    """Attempts to determine if a target file is binary."""
    # This is a bit false positive prone, as some "application" mime types are text.
    # However, as we only support a finite number of formats, we should be safe to do
    # this (for now!)
    binary_types = (
        "application",
        "image",
        "audio",
        "video",
    )
    if target.mime and target.mime.startswith(binary_types):
        return True

    # Otherwise, we'll try and read some data as text and see. This could fail if a
    # binary contained readable text for 10 * CHUNK_SIZE.
    try:
        with open(target.path, "r") as fin:
            for _ in range(0, 10):
                fin.read(CHUNK_SIZE)
    except UnicodeDecodeError:
        return True

    # Define to text.
    return False


def generate_sample(target: manifest.Entry, offset: int, size: int) -> finding.Sample:
    """Generates a sample for a finding."""
    binary = is_binary(target)

    before = bytes()
    after = bytes()
    entry = bytes()

    try:
        # Make sure we don't try and read past the beginning and end of the file.
        target_sz = os.stat(target.path).st_size

        if offset - WINDOW_SIZE < 0:
            before_sz = offset
            before_offset = 0
        else:
            before_sz = WINDOW_SIZE
            before_offset = offset - before_sz

        if offset + WINDOW_SIZE > target_sz:
            after_sz = target_sz - offset
            after_offset = after_sz
        else:
            after_sz = WINDOW_SIZE
            after_offset = offset + after_sz

        with open(target.path, "rb") as fin:
            # Seek to and read in the context before.
            fin.seek(before_offset)
            before = fin.read(before_sz)

            # Read the finding match itself. We have this already from yara, but we're
            # already here so we may as well.
            fin.seek(offset)
            entry = fin.read(size)

            # Seek to and read in the context after the finding.
            fin.seek(after_offset)
            after = fin.read(after_sz)
    except OSError as err:
        raise FileAccessException(err)

    if binary:
        return finding.Sample(
            window=WINDOW_SIZE,
            before=base64.b64encode(before),
            after=base64.b64encode(after),
            finding=base64.b64encode(entry),
            binary=binary,
        )
    else:
        return finding.Sample(
            window=WINDOW_SIZE,
            before=str(before, "utf-8"),
            after=str(after, "utf-8"),
            finding=str(entry, "utf-8"),
            binary=binary,
        )


def generate_location(target: manifest.Entry, offset: int) -> finding.Location:
    """Generates a location for a finding."""
    # If the file is binary, we can't generate a line number so we already have the data
    # we need.
    if is_binary(target):
        return finding.Location(offset=offset)

    # Attempt to generate a line number for the finding.
    bytes_read = 0
    line_number = 0
    try:
        with open(target.path, "r") as fin:
            # Read in chunks, counting the number of newline characters up to the chunk
            # which includes the finding.
            while bytes_read < offset:
                bytes_read += CHUNK_SIZE

                if bytes_read > offset:
                    line_number += fin.read(offset).count("\n")
                else:
                    line_number += fin.read(CHUNK_SIZE).count("\n")
    except OSError as err:
        raise FileAccessException(err)

    return finding.Location(offset=offset, line=line_number)


def generate_findings(target: manifest.Entry, match: yara.Match) -> List[finding.Entry]:
    """Attempts to create findings based on matches inside of the target file."""
    findings = []

    # Generate a new finding entry for each matched string. This is in order to ensure
    # that multiple findings in the same file are listed separately - as they may be
    # different credentials.
    for offset, _, entry in match.strings:
        location = generate_location(target, offset)
        sample = generate_sample(target, offset, len(entry))

        # Add on information about the origin of the finding (that's us!)
        source = finding.Source(
            module=__name__,
            reference=match.meta.get("name", "UNKNOWN"),
            version=match.meta.get("version", "UNKNOWN"),
            description=match.meta.get("description"),
        )
        findings.append(
            finding.Entry(
                md5=target.md5,
                path=target.overlay if target.overlay else target.path,
                confidence=match.meta.get("accuracy", 50),
                source=source,
                sample=sample,
                location=location,
            )
        )

    return findings


def matcher(target: manifest.Entry, ruleset: yara.Rules) -> List[finding.Entry]:
    findings = []

    for match in ruleset.match(target.path):
        findings.extend(generate_findings(target, match))

    return findings


def run(
    targets: List[manifest.Entry],
    pack: pack.Format,
    workers: int = 10,
    skip_on_eacces: bool = True,
) -> List[finding.Entry]:
    """
    Executes the rules based matcher on all input files, returning a list of finding
    Entry objects.
    """
    findings = []

    # Load and compile all YARA rules up front.
    namespaces = dict()

    for rule in pack.pack:
        namespace = hashlib.md5(bytes(rule.path, "utf-8")).hexdigest()
        namespaces[namespace] = rule.path

    try:
        ruleset = yara.compile(filepaths=namespaces)
    except yara.Error as err:
        raise InvalidFormatException(err)

    # Run the matcher in a thread pool as we're likely to be I/O bound.
    with ThreadPoolExecutor(max_workers=workers) as pool:
        futures = []

        # Reject any input files which are supported archives. This is as we should have
        # unpacked versions of these to process, which allows matching the specific file
        # with a finding, rather than a finding on an archive.
        #
        # NOTE: Credentials stuffed into metadata of supported archive formats which
        #       support archive metadata (such as Zip's "Extra") will not be found.
        #
        for target in targets:
            if target.mime not in archive.MIME_TYPE_HANDLERS:
                futures.append(pool.submit(matcher, target, ruleset))

        for future in as_completed(futures):
            try:
                findings.extend(future.result())
            except FileAccessException:
                if not skip_on_eacces:
                    raise

    return findings



"""Defines types to assist with loading and processing of manifests.

SPDX-License-Identifier: BSD-3-Clause
"""

from pydantic import BaseModel, Extra, Field


class Entry(BaseModel, extra=Extra.forbid):
    """Defines the schema of a file to process."""

    path: str = Field(
        None,
        title="The path to the file on disk.",
    )
    overlay: str = Field(
        None,
        title=(
            "The overlay path of a file. This is used to generate virtual paths which "
            "provider the path to files inside of archives."
        ),
    )
    md5: str = Field(
        None,
        title="The MD5 sum of the file.",
    )
    mime: str = Field(
        None,
        title="The mimetype of the file.",
    )


class Format(BaseModel, extra=Extra.forbid):
    """Defines the schema of a manifest file."""

    files: list[Entry] = Field(
        [],
        title="A list of files to scan.",
    )



"""STACS - Static Token And Credential Scanner.

SPDX-License-Identifier: BSD-3-Clause
"""

from stacs import __about__  # noqa:F401
from stacs import constants  # noqa:F401
from stacs import filter  # noqa:F401
from stacs import loader  # noqa:F401
from stacs import model  # noqa:F401
from stacs import output  # noqa:F401
from stacs import scanner  # noqa:F401



"""Defines a filter which sets the ignore flag on entries present in an ignore list.

SPDX-License-Identifier: BSD-3-Clause
"""

import re
from typing import List

from stacs.exceptions import IgnoreListException
from stacs.model import finding, ignore_list


def by_pattern(finding: finding.Entry, ignore: ignore_list.Entry) -> bool:
    """Process a regex ignore list entry."""
    # Short circuit if no pattern is set.
    if not ignore.pattern:
        return False

    # If there's a match on the path, check whether the ignore is for the same module.
    if re.search(ignore.pattern, finding.path):
        if ignore.module != finding.source.module:
            return False

        # Then check whether the ignore is for the particular reference.
        if ignore.references:
            if finding.source.reference in ignore.references:
                return True

            return False

        # Or check whether the ignore is for the same offest.
        if ignore.offset is not None:
            if finding.location.offset == ignore.offset:
                return True
            return False

        # In this case this is a fairly permissive ignore.
        return True

    return False


def by_path(finding: finding.Entry, ignore: ignore_list.Entry) -> bool:
    """Process a path based ignore list entry."""
    # Short circuit if no path is set.
    if not ignore.path:
        return False

    # If there's a match on the hash, check whether the ignore is for the same module.
    if ignore.path == finding.path:
        if finding.source.module != ignore.module:
            return False

        # Then check whether the ignore is for the particular reference.
        if ignore.references:
            if finding.source.reference in ignore.references:
                return True
            return False

        # Or check whether the ignore is for the same offest.
        if ignore.offset is not None:
            if finding.location.offset == ignore.offset:
                return True
            return False

        # In this case this is a fairly permissive ignore.
        return True

    return False


def by_hash(finding: finding.Entry, ignore: ignore_list.Entry) -> bool:
    """Process a hash based ignore list entry."""
    # Short circuit if no hash is set.
    if not ignore.md5:
        return False

    # If there's a match on the hash, check whether the ignore is for the same module.
    if ignore.md5 == finding.md5:
        if finding.source.module != ignore.module:
            return False

        # Then check whether the ignore is for the particular reference.
        if ignore.references:
            if finding.source.reference in ignore.references:
                return True
            return False

        # Or check whether the ignore is for the same offest.
        if ignore.offset is not None:
            if finding.location.offset == ignore.offset:
                return True
            return False

        # In this case this is a fairly permissive ignore.
        return True

    return False


def process(
    findings: List[finding.Entry],
    ignore_list: ignore_list.Format,
) -> List[finding.Entry]:
    """Processes an ignore list and marks the relevant findings as ignored."""
    filtered_findings = []

    for entry in findings:
        for ignore in ignore_list.ignore:
            try:
                if by_path(entry, ignore):
                    ignore = finding.Ignore(
                        ignored=True,
                        reason=ignore.reason,
                    )
                    entry.ignore = ignore
                    break

                if by_pattern(entry, ignore):
                    ignore = finding.Ignore(
                        ignored=True,
                        reason=ignore.reason,
                    )
                    entry.ignore = ignore
                    break

                if by_hash(entry, ignore):
                    ignore = finding.Ignore(
                        ignored=True,
                        reason=ignore.reason,
                    )
                    entry.ignore = ignore
                    break
            except re.error as err:
                raise IgnoreListException(
                    f"Error in ignore list entry '{ignore.reason}': {err}"
                )

        # Add the finding to our results, whether updated or not.
        filtered_findings.append(entry)

    return filtered_findings



"""Defines handlers for unpacking of archives.

SPDX-License-Identifier: BSD-3-Clause
"""

import bz2
import gzip
import hashlib
import lzma
import os
import shutil
import tarfile
import zipfile

from stacs.constants import CHUNK_SIZE
from stacs.exceptions import FileAccessException, InvalidFileException


def path_hash(filepath: str) -> str:
    """Returns a hash of the filepath, for use with unique directory creation."""
    return hashlib.md5(bytes(filepath, "utf-8")).hexdigest()


def zip_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided archive."""
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    # Attempt to unpack the zipfile to the new unpack directory.
    try:
        with zipfile.ZipFile(filepath, "r") as archive:
            archive.extractall(directory)
    except zipfile.BadZipFile as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {directory}: {err}"
        )


def tar_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided archive."""
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    # Attempt to unpack the tarball to the new unpack directory.
    try:
        with tarfile.open(filepath, "r") as archive:
            archive.extractall(directory)
    except tarfile.TarError as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {directory}: {err}"
        )


def gzip_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided archive."""
    output = ".".join(os.path.basename(filepath).split(".")[:-1])

    # Ensure that files with a proceeding dot are properly handled.
    if len(output) < 1:
        output = os.path.basename(filepath).lstrip(".")

    # Although gzip files cannot contain more than one file, we'll still spool into
    # a subdirectory under the cache for consistency.
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    # TODO: This can likely be optimized for tgz files, as currently the file will be
    #       first processed and gunzipped, and then reprocessed to be extracted.
    try:
        with gzip.open(filepath, "rb") as fin:
            with open(os.path.join(directory, output), "wb") as fout:
                shutil.copyfileobj(fin, fout, CHUNK_SIZE)
    except gzip.BadGzipFile as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {output}: {err}"
        )


def bzip2_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided archive."""
    output = ".".join(os.path.basename(filepath).split(".")[:-1])

    # Like gzip, bzip2 cannot support more than a single file. Again, we'll spool into
    # a subdirectory for consistency.
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    # TODO: This can likely be optimized for tbz files, as currently the file will be
    #       first processed and gunzipped, and then reprocessed to be extracted.
    try:
        with bz2.open(filepath, "rb") as fin:
            with open(os.path.join(directory, output), "wb") as fout:
                shutil.copyfileobj(fin, fout, CHUNK_SIZE)
    except ValueError as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {output}: {err}"
        )


def lzma_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided archive."""
    output = ".".join(os.path.basename(filepath).split(".")[:-1])

    # Ensure that files with a proceeding dot are properly handled.
    if len(output) < 1:
        output = os.path.basename(filepath).lstrip(".")

    # Although xz files cannot contain more than one file, we'll still spool into
    # a subdirectory under the cache for consistency.
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    try:
        with lzma.open(filepath, "rb") as fin:
            with open(os.path.join(directory, output), "wb") as fout:
                shutil.copyfileobj(fin, fout, CHUNK_SIZE)
    except lzma.LZMAError as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {output}: {err}"
        )


def get_mimetype(chunk: bytes) -> str:
    """Attempts to locate the appropriate handler for a given file.

    This may fail if the required "magic" is at an offset greater than the CHUNK_SIZE.
    However, currently this is not an issue, but may need to be revisited later as more
    archive types are supported.
    """
    for name, options in MIME_TYPE_HANDLERS.items():
        offset = options["offset"]
        magic = options["magic"]

        for candidate in magic:
            if chunk[offset : (offset + len(candidate))] == candidate:  # noqa: E203
                return name

    return None


# Define all supported archives and their handlers. As we currently only support a small
# list of types we can just define file magic directly here, rather than use an external
# library. This removes the need for dependencies which may have other system
# dependencies - such as libmagic. It should also provide a small a speed up during
# unpacking, as we're only looking for a small number of types.
MIME_TYPE_HANDLERS = {
    "application/x-tar": {
        "offset": 257,
        "magic": [
            bytearray([0x75, 0x73, 0x74, 0x61, 0x72]),
        ],
        "handler": tar_handler,
    },
    "application/gzip": {
        "offset": 0,
        "magic": [
            bytearray([0x1F, 0x8B]),
        ],
        "handler": gzip_handler,
    },
    "application/x-bzip2": {
        "offset": 0,
        "magic": [
            bytearray([0x42, 0x5A, 0x68]),
        ],
        "handler": bzip2_handler,
    },
    "application/zip": {
        "offset": 0,
        "magic": [
            bytearray([0x50, 0x4B, 0x03, 0x04]),
            bytearray([0x50, 0x4B, 0x05, 0x06]),
            bytearray([0x50, 0x4B, 0x07, 0x08]),
        ],
        "handler": zip_handler,
    },
    "application/x-xz": {
        "offset": 0,
        "magic": [
            bytearray([0xFD, 0x37, 0x7A, 0x58, 0x5A, 0x00]),
        ],
        "handler": lzma_handler,
    },
}



"""Defines models used by STACS.

SPDX-License-Identifier: BSD-3-Clause
"""

from stacs.model import finding  # noqa: F401
from stacs.model import ignore_list  # noqa: F401
from stacs.model import manifest  # noqa: F401
from stacs.model import pack  # noqa: F401



"""Defines a file path loader for STACS.

SPDX-License-Identifier: BSD-3-Clause
"""

import hashlib
import logging
import os
import re
import shutil
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List

from stacs.constants import ARCHIVE_FILE_SEPARATOR, CHUNK_SIZE
from stacs.exceptions import FileAccessException
from stacs.loader import archive
from stacs.model.manifest import Entry

logger = logging.getLogger(__name__)


def metadata(filepath: str, overlay: str = None) -> Entry:
    """Generates a hash and determines the mimetype of the input file."""
    md5 = hashlib.md5()
    mime = None

    # Read the file in chunks.
    try:
        with open(filepath, "rb") as fin:
            while chunk := fin.read(CHUNK_SIZE):
                md5.update(chunk)

                # Only attempt to determine the filetype on the first chunk.
                if not mime and fin.tell() <= CHUNK_SIZE:
                    mime = archive.get_mimetype(chunk)
    except OSError as err:
        raise FileAccessException(f"Unable to open file at {filepath}: {err}")

    return Entry(
        path=filepath,
        md5=md5.hexdigest(),
        mime=mime,
        overlay=overlay,
    )


def walker(path: str, skip_on_eacces: bool) -> List[str]:
    """Recursively walk a file path, returning a list of all files."""
    entries = []

    # TODO: Would moving walker to a generator yield a performance increase, or lead to
    #       higher disk contention due to the hasher running at the same time?
    try:
        with os.scandir(path) as scan:
            for handle in scan:
                try:
                    # Recurse on directories, but not symlinks.
                    if handle.is_dir() and not handle.is_symlink():
                        entries.extend(walker(handle.path, skip_on_eacces))

                    # Track files, but not symlinks.
                    if handle.is_file() and not handle.is_symlink():
                        entries.append(handle.path)
                except PermissionError:
                    if not skip_on_eacces:
                        raise
                except OSError:
                    # This is usually due to too many levels of symlinks. However, other
                    # cases are likely with a large enough input.
                    continue
    except NotADirectoryError:
        entries.append(path)

    return list(set(entries))


def qualify(path: str) -> str:
    """Add the scheme to a file path, if required."""
    if path.startswith("/"):
        return f"file://{path}"
    else:
        return path


def finder(
    path: str,
    cache: str,
    workers: int = 10,
    skip_on_eacces: bool = True,
) -> List[Entry]:
    """Processes the input path, returning a list of all files and their hashes."""
    entries = []
    futures = dict()

    # Run the metadata enumerator in a thread pool as we're likely to be I/O bound.
    with ThreadPoolExecutor(max_workers=workers) as pool:
        futures = {
            pool.submit(metadata, file): file for file in walker(path, skip_on_eacces)
        }

        # A loop and counter is used here to ensure that additional work which may be
        # submitted during the 'final loop' isn't accidentally ignored.
        while True:
            complete = 0
            for future in as_completed(futures):
                complete += 1

                try:
                    result = future.result()
                except FileAccessException:
                    if not skip_on_eacces:
                        raise

                # Track the result and then remove the future from the initial futures
                # list so that these results aren't returned again next iteration.
                entries.append(result)
                del futures[future]

                # Check it the file was found to be an archive, and if so, unpack it.
                handler = archive.MIME_TYPE_HANDLERS.get(result.mime, {}).get("handler")
                if not handler:
                    continue

                # Remove any existing previously unpacked files, then unpack the archive
                # and submit extracted files back into the queue. This is to allow for
                # easy recursive unpacking of nested archives.
                destination = os.path.join(cache, archive.path_hash(result.path))
                shutil.rmtree(destination, ignore_errors=True)

                handler(result.path, destination)
                for file in walker(destination, skip_on_eacces):
                    # The overlay path is a 'virtual' path that is constructed based on
                    # the archive the file appears inside of, and the path of the file
                    # inside of the archive. However, as archives may be nested, we need
                    # to check whether we already have an overlay and, if set, use that
                    # value instead.
                    if result.overlay:
                        parent = result.overlay
                    else:
                        parent = result.path

                    overlay = (
                        f"{parent}"
                        f"{ARCHIVE_FILE_SEPARATOR}"
                        f"{re.sub(rf'^{destination}/?', '', file)}"
                    )

                    # Submit back to the pool for processing.
                    submission = pool.submit(metadata, file, overlay=overlay)
                    futures[submission] = file

            if complete == 0:
                break

    return entries



"""Defines loaders used by STACS.

SPDX-License-Identifier: BSD-3-Clause
"""

from stacs.loader import archive  # noqa:F401
from stacs.loader import filepath  # noqa:F401
from stacs.loader import manifest  # noqa:F401



"""Defines types to assist with reporting findings.

SPDX-License-Identifier: BSD-3-Clause
"""

from pydantic import BaseModel, Extra, Field


class Location(BaseModel, extra=Extra.forbid):
    """Defines data associated with a location of a finding."""

    line: int = Field(
        None,
        title="The line number which contains the finding.",
    )
    offset: int = Field(
        None,
        title="The offset from the start of the file of the finding (in bytes).",
    )


class Source(BaseModel, extra=Extra.forbid):
    """Defines data associated with the source of a finding."""

    module: str = Field(
        title="The STACS module which generated the finding.",
    )
    description: str = Field(
        None,
        title="A description of the finding",
    )
    reference: str = Field(
        title="A reference to the element which generated the finding."
    )
    version: str = Field(
        None,
        title="The version of the element which generated the finding.",
    )


class Sample(BaseModel, extra=Extra.forbid):
    """The content and context of a finding."""

    window: int = Field(
        title="The number of bytes before and after a finding included in the sample.",
    )
    before: str = Field(
        title="The contents of N bytes before the finding.",
    )
    after: str = Field(
        title="The contents of N bytes after the finding.",
    )
    finding: str = Field(
        title="The contents of the finding.",
    )
    binary: bool = Field(
        title="Indicates that the finding was binary and is base64 encoded."
    )


class Ignore(BaseModel, extra=Extra.forbid):
    """Defines the ignore schema of a finding."""

    ignored: bool = Field(
        False,
        title="Whether the finding should be ignored due to allow list.",
    )
    reason: str = Field(
        title="The reason to ignore the finding.",
    )


class Entry(BaseModel, extra=Extra.forbid):
    """Defines the schema of a finding."""

    path: str = Field(
        title="The path to the file.",
    )
    md5: str = Field(
        title="The MD5 sum of the file.",
    )
    confidence: float = Field(
        None,
        title="The confidence of the finding.",
    )
    location: Location = Field(
        None,
        title="The location of the finding in the input file.",
    )
    sample: Sample = Field(
        None,
        title="Information relating to the content of the finding.",
    )
    source: Source = Field(
        None,
        title="Information about the source of the finding.",
    )
    ignore: Ignore = Field(
        None,
        title="Information about whether the entry should be ignored.",
    )



"""Defines types to assist with loading and processing of ignore lists.

SPDX-License-Identifier: BSD-3-Clause
"""

import json
import os
from typing import List

from pydantic import BaseModel, Extra, Field, validator
from stacs.exceptions import IgnoreListException, STACSException


class Entry(BaseModel, extra=Extra.forbid):
    """Defines the schema of an ignore."""

    path: str = Field(
        None,
        title="The path of a file to ignore.",
    )
    pattern: str = Field(
        None,
        title="A pattern of the file path to ignore.",
    )
    reason: str = Field(
        title="The reason for ignoring the finding.",
    )
    md5: str = Field(
        None,
        title="The MD5 sum of the file to ignore.",
    )
    module: str = Field(
        "stacs.scanner.rules",
        title="Which module to ignore findings from.",
    )
    references: List[str] = Field(
        [],
        title=(
            "A list of references to ignore findings from, defaults to all if not set."
        ),
    )
    offset: int = Field(
        None,
        title="The offset of the specific finding to ignore.",
    )

    @validator("path", always=True)
    def exclusive_path_or_pattern(cls, value, values):
        """Ensure that either path or pattern is provided, not both."""
        if values.get("pattern") and value:
            raise IgnoreListException(
                "Either path OR pattern must be specified, not both."
            )

        if values.get("pattern") and not value and not values.get("md5"):
            raise IgnoreListException("One of pattern, path, or md5 must be set.")

        return value

    @validator("offset", "references", always=True)
    def offset_and_references_requires_module(cls, value, values):
        """Ensure that if offset or references is set, module is as well."""
        if not values.get("module"):
            raise IgnoreListException("Module must be set for this type of ignore.")

        return value

    @validator("offset", always=True)
    def offset_and_refernces_both_set(cls, value, values):
        if value and len(values.get("references")) > 0:
            raise IgnoreListException(
                "An offset cannot be combined with a list of references."
            )

        return value


class Format(BaseModel, extra=Extra.forbid):
    """Defines the schema of the ignore list."""

    include: List[str] = Field(
        [],
        title="Define a list of additional ignore lists to include.",
    )
    ignore: List[Entry] = Field(
        [],
        title="Define a list of ignore list entries.",
    )


def from_file(filename: str) -> Format:
    """Load an ignore list from file, returning a rendered down and complete list."""
    parent_file = os.path.abspath(os.path.expanduser(filename))
    parent_path = os.path.dirname(os.path.abspath(os.path.expanduser(filename)))

    # Load the parent ignore list, and then recurse as needed to handle includes.
    try:
        with open(parent_file, "r") as fin:
            parent_list = Format(**json.load(fin))

        # Roll over the ignore list and replace all entries with a fully qualified path,
        # if not already set.
        for index, path in enumerate(parent_list.include):
            if not path.startswith("/"):
                del parent_list.include[index]
                parent_list.include.insert(index, os.path.join(parent_path, path))
    except (OSError, json.JSONDecodeError) as err:
        raise STACSException(err)

    # Recursively load included ignore lists.
    for file in parent_list.include:
        child_pack = from_file(os.path.join(parent_path, file))
        parent_list.ignore.extend(child_pack.ignore)

    # Finally strip the included ignore lists from the entry, as these have been
    # resolved, returning the loaded ignore lists to the caller.
    parent_list.include = []
    return parent_list



"""Define constants commonly used throughout STACS.

SPDX-License-Identifier: BSD-3-Clause
"""

# The size of chunks to use when reading files.
CHUNK_SIZE = 65536

# The size, in bytes, of the sample window.
WINDOW_SIZE = 20

# Define the default cache directory, used to unpack archives into.
CACHE_DIRECTORY = "/tmp"

# Define the character to use when constructed paths to findings which are inside of
# archives.
ARCHIVE_FILE_SEPARATOR = "!"



"""STACS Exceptions.

SPDX-License-Identifier: BSD-3-Clause
"""


class STACSException(Exception):
    """The most generic form of exception raised by STACS."""


class FileAccessException(STACSException):
    """Indicates an error occured while attempting to access a file."""


class InvalidFileException(STACSException):
    """Indicates the format of a file did not match what was expected."""


class InvalidFormatException(STACSException):
    """Indicates that the format of a rule did not match what was expected."""


class IgnoreListException(STACSException):
    """Indicates an invalid ignore list was provided."""


class NotImplementedException(STACSException):
    """Indicates that the requested method has not been implemented."""



"""Defines types to assist with loading and processing of rule packs.

SPDX-License-Identifier: BSD-3-Clause
"""

import json
import os
from typing import List

from pydantic import BaseModel, Extra, Field
from stacs.exceptions import STACSException


class Entry(BaseModel, extra=Extra.forbid):
    """Defines the schema of an allow."""

    module: str = Field(
        "rules",
        title="Which module the rules are for.",
    )
    path: str = Field(
        None,
        title="The path of a the module's rule to load.",
    )


class Format(BaseModel, extra=Extra.forbid):
    """Defines the schema of the rule pack."""

    include: List[str] = Field(
        [],
        title="Define a list of additional packs to include.",
    )
    pack: List[Entry] = Field(
        [],
        title="A list of pack entries.",
    )


def from_file(filename: str) -> Format:
    """Load a pack from file, returning a rendered down and complete pack."""
    parent_file = os.path.abspath(os.path.expanduser(filename))
    parent_path = os.path.dirname(os.path.abspath(os.path.expanduser(filename)))

    # Load the parent pack, and then recurse as needed to handle includes.
    try:
        with open(parent_file, "r") as fin:
            parent_pack = Format(**json.load(fin))

        # Roll over the pack and ensure any entries are fully qualified.
        for index, entry in enumerate(parent_pack.pack):
            if not entry.path.startswith("/"):
                # Resolve the path.
                qualified = Entry(
                    path=os.path.join(parent_path, entry.path),
                    module=entry.module,
                )

                # Swap the entry with one that's fully qualified.
                del parent_pack.pack[index]
                parent_pack.pack.insert(index, qualified)
    except (OSError, json.JSONDecodeError) as err:
        raise STACSException(err)

    # Recursively load included packs, adding results to the loaded pack.
    for file in parent_pack.include:
        child_pack = from_file(os.path.join(parent_path, file))
        parent_pack.pack.extend(child_pack.pack)

    # Finally strip the included packs from the entry, as these have been resolved,
    # returning the loaded pack to the caller.
    parent_pack.include = []
    return parent_pack



"""STACS Setup.

SPDX-License-Identifier: BSD-3-Clause
"""

from setuptools import setup

# Defer to setupmeta.
setup(
    name="stacs",
    setup_requires="setupmeta",
    description="Static Token And Credential Scanner.",
)



"""STACS - Static Token And Credential Scanner.

SPDX-License-Identifier: BSD-3-Clause
"""

__title__ = "stacs"
__summary__ = "Static Token And Credential Scanner."
__version__ = "0.1.1"
__author__ = "Peter Adkins"
__uri__ = "https://www.github.com/stacscan/stacs/"
__license__ = "BSD-3-Clause"



"""Defines a markdown output handler for STACS.

SPDX-License-Identifier: BSD-3-Clause
"""

from typing import List

from stacs import model
from stacs.exceptions import NotImplementedException


def render(findings: List[model.finding.Entry], pack: model.pack.Format) -> str:
    return NotImplementedException("Markdown output not yet implemented, sorry!")



"""Defines the primary STACS CLI entrypoint.

SPDX-License-Identifier: BSD-3-Clause
"""

import logging
import os
import sys

import click
import stacs


@click.command()
@click.version_option()
@click.option(
    "--debug",
    is_flag=True,
    help="Increase verbosity of logs for debugging",
)
@click.option(
    "--rule-pack",
    help="The path to the rule pack to load.",
    default="~/.stacs/pack.json",
)
@click.option(
    "--ignore-list",
    help="The path to the ignore list to load (if required).",
)
@click.option(
    "--cache-directory",
    help="The path to use as a cache - used when unpacking archives.",
    default=stacs.constants.CACHE_DIRECTORY,
)
@click.argument("path")
def main(
    debug: bool,
    rule_pack: str,
    ignore_list: str,
    cache_directory: str,
    path: str,
) -> None:
    """STACS - Static Token And Credential Scanner."""
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.INFO,
        format="%(asctime)s - %(process)d - [%(levelname)s] %(message)s",
    )
    logger = logging.getLogger("stacs")

    # Expand the input path.
    path = os.path.abspath(os.path.expanduser(path))

    # Load the rule pack.
    logger.info(f"Attempting to load rule pack from {rule_pack}")
    try:
        pack = stacs.model.pack.from_file(rule_pack)
    except stacs.exceptions.STACSException as err:
        logger.error(f"Unable to load rule pack: {err}")
        sys.exit(-1)

    # Load the ignore list.
    ignored = []
    if ignore_list:
        logger.info(f"Attempting to load ignore list from {ignore_list}")
        try:
            ignored = stacs.model.ignore_list.from_file(ignore_list)
        except stacs.exceptions.STACSException as err:
            logger.error(f"Unable to load ignore list: {err}")
            sys.exit(-1)

    # Generate a list of candidate files to scan.
    logger.info(f"Attempting to get a list of files to scan from {path}")
    try:
        targets = stacs.loader.filepath.finder(path, cache_directory)
    except stacs.exceptions.STACSException as err:
        logger.error(f"Unable to generate file list: {err}")
        sys.exit(-2)

    # Submit files for analysis.
    logger.info(f"Found {len(targets)} files for analysis")

    findings = []
    for scanner in stacs.scanner.__all__:
        try:
            findings.extend(getattr(stacs.scanner, scanner).run(targets, pack))
        except stacs.exceptions.InvalidFormatException as err:
            logger.error(f"Unable to load a rule in scanner {scanner}: {err}")
            continue

    # Filter findings by allow list.
    if ignored:
        findings = stacs.filter.ignore_list.process(findings, ignored)

    # Generate SARIF and output to STDOUT.
    logger.info("Generaing SARIF from findings")
    try:
        sarif = stacs.output.sarif.render(path, findings, pack)
    except stacs.exceptions.STACSException as err:
        logger.error(f"Unable to generate SARIF: {err}")
        sys.exit(-3)

    # TODO: Add file output as an option.
    logger.info(f"Found {len(findings)} findings")
    print(sarif)
